// DeciderTiler.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "DeciderTiler.h"
#include "counters.h"

#define ACCEPTCOLOR "RGB(255,0,0)"
#define REJECTCOLOR "RGB(100,100,100)"


int GetCounterBit(std::string cv)
{
  for (size_t i=0; i<cv.length(); i++) {
    if (cv.at(i) == '0') {
      return 0;
    } else if (cv.at(i) == '1') {
      return 1;
    }
  }
  std::string se = "ERROR: Illegal value '" + cv + " in GetCounterBit";
  printf(se.c_str());
  exit(1);
}

void Usage()
{
  printf("\nUsage:  DeciderTiler.exe -tm <TM definition file> -out <tileset output file> [-seed <seed output file>] [-norotate] [-quadopt]");
  printf("\n        DeciderTiler.exe -tmpl <template output file>");
  printf("\n");
  printf("\n  -tm      : name of the input file to be used as the Turing machine");
  printf("\n             specification");
  printf("\n  -out     : name of the output file for the generated tileset");
  printf("\n  -seed    : [optional] name of the output file for the generated tile seed");
  printf("\n  -norotate: [optional] if this option is present, the tile types will not be");
  printf("\n             rotated so that the solutions are placed on the positive x-axis,");
  printf("\n             but instead will be on the negative y-axis and the assembly will");
  printf("\n             essentially grow upward");
  printf("\n  -quadopt : [optional] generate a tileset that assembles in a quadrant optimal");
  printf("\n             manner (assuming L(M) requires sufficient space complexity)");
  printf("\n  -tmpl    : name of the output file to be created with a template for a Turing");
  printf("\n             machine specification");
  printf("\n");
  exit(0);
}

int main(int argc, char* argv[])
{
  std::string TMDefFile;
  std::string TileOutFile;
  std::string SeedOutFile;
  bool bDoRotate    = true;
  bool bQuadOptimal = false;

  for (int i=1; i<argc; i++) {
    char* foo = argv[i];
    if (strcmp(argv[i], "-tm") == 0) {
      if (argc >= i+1) {
        TMDefFile = argv[++i];
      } else {
        Usage();
      }
    } else if (strcmp(argv[i], "-out") == 0) {
      if (argc >= i+1) {
        TileOutFile = argv[++i];
      } else {
        Usage();
      }
    } else if (strcmp(argv[i], "-seed") == 0) {
      if (argc >= i+1) {
        SeedOutFile = argv[++i];
      } else {
        Usage();
      }
    } else if (strcmp(argv[i], "-tmpl") == 0) {
      if (argc >= i+1) {
        PrintTemplate(argv[++i]);
        exit(0);
      } else {
        Usage();
      }
    } else if (strcmp(argv[i], "-norotate") == 0) {
      bDoRotate = false;
    } else if (strcmp(argv[i], "-quadopt") == 0) {
      bQuadOptimal = true;
    }
  }
  if ((TMDefFile.length() <= 0) || (TileOutFile.length() <= 0)) {
    Usage();
  }

  if (bQuadOptimal) {
    bDoRotate = false;
  }

  //////////////////////////////////////////////////////////////////////
  //  Read the Turning machine defintion from the input file

  printf("\nProcessing Turing Machine definition file: '%s'...\n", TMDefFile.c_str());

  std::ifstream fileIn(TMDefFile.c_str());
  if (!fileIn.is_open()) {
    printf("Error:  Unable to open input file '%s'\n", TMDefFile.c_str());
    exit(1);  
  }

  TuringMachine TM;
  TM.m_alphabet.push_back("0");
  TM.m_alphabet.push_back("1");

  std::string   seedStr;

  char lineStr[1001] = {0};
  char *pos = NULL;
  bool bInTransitions = false;
  while (!fileIn.eof()) {
    fileIn.getline(lineStr, 1000);
    std::string tempLine(lineStr);
    tempLine = StripSpaces(tempLine);
    if ((tempLine.length() > 0) && (strncmp(lineStr, "//", 2) != 0)) {    // Skip comment lines and empty lines
      std::vector<std::string> arr;
      if (strncmp(lineStr, STATES_STR, strlen(STATES_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(STATES_STR);
        std::string sLine(pos);
        SplitString(sLine, std::string(","), arr);
        for (size_t i=0; i<arr.size(); i++) {
          TM.m_states.push_back(StripSpaces(arr[i]).c_str());
        }
      } else if (strncmp(lineStr, HALTING_STR, strlen(HALTING_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(HALTING_STR);
        std::string sLine(pos);
        SplitString(sLine, std::string(","), arr);
        TM.m_haltingState = StripSpaces(arr[0]).c_str();
      } else if (strncmp(lineStr, ACCEPTING_STR, strlen(ACCEPTING_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(ACCEPTING_STR);
        std::string sLine(pos);
        SplitString(sLine, std::string(","), arr);
        TM.m_acceptingState = StripSpaces(arr[0]).c_str();
      } else if (strncmp(lineStr, REJECTING_STR, strlen(REJECTING_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(REJECTING_STR);
        std::string sLine(pos);
        SplitString(sLine, std::string(","), arr);
        TM.m_rejectingState = StripSpaces(arr[0]).c_str();
      } else if (strncmp(lineStr, ALPHABET_STR, strlen(ALPHABET_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(ALPHABET_STR);
        std::string sLine(pos);
        SplitString(sLine, std::string(","), arr);
        for (size_t i=0; i<arr.size(); i++) {
          std::string s = StripSpaces(arr[i]).c_str();
          if ((s == "0") || (s == "1")) {
            printf("\nWarning:  The symbols '0' and '1' make up the input alphabet and are therefore included in the tape alphabet by default.\n\n");
          } else if (s == "_") {
            printf("\nError:  The symbol '_' is a reserved character for a blank tape symbol and is included in the tape alphabet which is automatically generated as the input alphabet plus that symbol.  It cannot also be included in the input alphabet.\n\n");
            fileIn.close();
            exit(1);
          } else {
            TM.m_alphabet.push_back(s);
          }
        }
        TM.m_alphabet.push_back(BLANK);
      } else if (strncmp(lineStr, START_STR, strlen(START_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(START_STR);
        std::string sLine(pos);
        TM.m_startState = StripSpaces(sLine);
      } else if (strncmp(lineStr, TRANSITION_STR, strlen(TRANSITION_STR)) == 0) {
        pos = lineStr + strlen(TRANSITION_STR);
        bInTransitions = true;
      } else if (bInTransitions) {
        Transition tran;
        std::string sLine = lineStr;
        size_t mid = sLine.find("->");

        if (mid == -1) {
          printf("\nError:  Invalid transition specification: '%s'\n\n", sLine.c_str());
          fileIn.close();
          exit(1);
        }

        std::string left = sLine.substr(0, mid);
        SplitString(left, std::string(","), arr);
        tran.m_startState = StripSpaces(arr[0]).c_str();
        tran.m_input = StripSpaces(arr[1]);
        arr.clear();

        std::string right = sLine.substr(mid+2, sLine.length());
        SplitString(right, std::string(","), arr);
        tran.m_endState = StripSpaces(arr[0]).c_str();
        tran.m_output = StripSpaces(arr[1]).c_str();
        std::string sDir = StripSpaces(arr[2]);
        if (sDir.compare("r")==0 || sDir.compare("R")==0) {
          tran.m_direction = RIGHT;
        } else if (sDir.compare("l")==0 || sDir.compare("L")==0) {
          tran.m_direction = LEFT;
        } else {
          tran.m_direction = INVALID;
        }

        if (tran.IsValid()) {
          TM.m_transitions.push_back(tran);
        } else {
          printf("Error:  Invalid transition: %s\n", sLine);
        }
      }
    }
  }
  fileIn.close();

  //////////////////////////////////////////////////////////////////////
  //  Verification of Turing machine definition
  if (TM.m_acceptingState.length() == 0) {
    printf("Error:  No accepting state specified for Turing machine!\n");
    exit(1);
  } else if (TM.m_rejectingState.length() == 0) {
    printf("Error:  No rejecting state specified for Turing machine!\n");
    exit(1);
  } else if (TM.m_startState.length() == 0) {
    printf("Error:  No start state specified for Turing machine!\n");
    exit(1);
  }

  bool bStartFound  = false;
  bool bAcceptFound = false;
  bool bRejectFound = false;
  for (StateSet::iterator it = TM.m_states.begin(); it != TM.m_states.end(); it++) {
    if (*it == TM.m_startState) {
      bStartFound = true;
    }
    if (*it == TM.m_acceptingState) {
      bAcceptFound = true;
    }
    if (*it == TM.m_rejectingState) {
      bRejectFound = true;
    }
  }
  if (!bStartFound) {
    printf("Error:  Start state is not defined in the set of states!\n");
    exit(1);
  }
  if (!bAcceptFound) {
    printf("Error:  Accepting state is not defined in the set of states!\n");
    exit(1);
  }
  if (!bRejectFound) {
    printf("Error:  Rejecting state is not defined in the set of states!\n");
    exit(1);
  }

  if (!TM.IsValid()) {
    printf("Error:  Invalid Turing machine definition!\n");
    exit(1);
  }

  TileSet tsTM;


  //////////////////////////////////////////////////////////////////////
  //  Create the equivalent tile set from the Turing machine's definition

  //  Create tiles for:
  //  1.  copy

  for (size_t k=0; k<TM.m_alphabet.size(); k++) {
    Tile a;
    a.m_display                 = TM.m_alphabet[k];
    a.m_sides[NORTH].m_strength = 1;
    a.m_sides[EAST].m_strength  = 1;
    a.m_sides[SOUTH].m_strength = 1;
    a.m_sides[WEST].m_strength  = 1;
    a.m_sides[NORTH].m_label    = TM.m_alphabet[k];
    a.m_sides[EAST].m_label     = COPYLEFT;
    a.m_sides[SOUTH].m_label    = TM.m_alphabet[k];
    a.m_sides[WEST].m_label     = COPYLEFT;
    tsTM.push_back(a);

    Tile a2;
    a2.m_display                 = TM.m_alphabet[k];
    a2.m_sides[NORTH].m_strength = 1;
    a2.m_sides[EAST].m_strength  = 1;
    a2.m_sides[SOUTH].m_strength = 1;
    a2.m_sides[WEST].m_strength  = 1;
    a2.m_sides[NORTH].m_label    = TM.m_alphabet[k];
    a2.m_sides[EAST].m_label     = COPYRIGHT;
    a2.m_sides[SOUTH].m_label    = TM.m_alphabet[k];
    a2.m_sides[WEST].m_label     = COPYRIGHT;
    tsTM.push_back(a2);
  }

  for (size_t t=0; t<TM.m_transitions.size(); t++) {
    std::string endState = TM.m_transitions[t].m_endState;

    for (size_t t2=0; t2<TM.m_transitions.size(); t2++) {
      std::string startState = TM.m_transitions[t2].m_startState;

      if (startState == endState) {

        std::string northLabel = "[" + endState + "]," + TM.m_transitions[t2].m_input;

        for (size_t t3=0; t3<TM.m_transitions.size(); t3++) {
          if ((TM.m_transitions[t3].m_startState == endState) &&
              (TM.m_transitions[t3].m_input == TM.m_transitions[t2].m_input))
          {
            if (TM.m_transitions[t3].m_endState == TM.m_acceptingState) {
              northLabel = ACCEPT;
              break;
            } else if (TM.m_transitions[t3].m_endState == TM.m_rejectingState) {
              northLabel = REJECT;
              break;
            }
          }
        }

        if (TM.m_transitions[t].m_direction == LEFT) {
          Tile m;
          m.m_display                 = "[" + endState + "]," + TM.m_transitions[t2].m_input;
          m.m_sides[NORTH].m_strength = 2;
          m.m_sides[EAST].m_strength  = 1;
          m.m_sides[SOUTH].m_strength = 1;
          m.m_sides[WEST].m_strength  = 1;
          m.m_sides[NORTH].m_label    = northLabel;
          m.m_sides[EAST].m_label     = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
          m.m_sides[SOUTH].m_label    = TM.m_transitions[t2].m_input;
          m.m_sides[WEST].m_label     = COPYLEFT;
          tsTM.push_back(m); 
        } else {
          Tile m;
          m.m_display                 = "[" + endState + "]," + TM.m_transitions[t2].m_input;
          m.m_sides[NORTH].m_strength = 2;
          m.m_sides[EAST].m_strength  = 1;
          m.m_sides[SOUTH].m_strength = 1;
          m.m_sides[WEST].m_strength  = 1;
          m.m_sides[NORTH].m_label    = northLabel;
          m.m_sides[EAST].m_label     = COPYRIGHT;
          m.m_sides[SOUTH].m_label    = TM.m_transitions[t2].m_input;
          m.m_sides[WEST].m_label     = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
          tsTM.push_back(m);
        }
      }
    }
    
    if (TM.m_transitions[t].m_direction == LEFT) {
      Tile top;
      top.m_display                 = TM.m_transitions[t].m_output;
      top.m_sides[NORTH].m_strength = 1;
      top.m_sides[EAST].m_strength  = 1;
      top.m_sides[SOUTH].m_strength = 2;
      top.m_sides[WEST].m_strength  = 1;
      top.m_sides[NORTH].m_label    = TM.m_transitions[t].m_output;
      top.m_sides[EAST].m_label     = COPYRIGHT;
      top.m_sides[SOUTH].m_label    = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
      top.m_sides[WEST].m_label     = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
      tsTM.push_back(top);
    } else {
      Tile top;
      top.m_display                 = TM.m_transitions[t].m_output;
      top.m_sides[NORTH].m_strength = 1;
      top.m_sides[EAST].m_strength  = 1;
      top.m_sides[SOUTH].m_strength = 2;
      top.m_sides[WEST].m_strength  = 1;
      top.m_sides[NORTH].m_label    = TM.m_transitions[t].m_output;
      top.m_sides[EAST].m_label     = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
      top.m_sides[SOUTH].m_label    = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
      top.m_sides[WEST].m_label     = COPYLEFT;
      tsTM.push_back(top);
    }
  }

  //  Create the main tileset to be output
  TileSet ts;

  //  Right side ('space') tiles
  Tile right2("RIGHT1");
  right2.m_display                  = BLANK;
  right2.m_sides[NORTH].m_strength  = 1;
  right2.m_sides[EAST].m_strength   = 2;
  right2.m_sides[SOUTH].m_strength  = 1;
  right2.m_sides[WEST].m_strength   = 1;
  right2.m_sides[NORTH].m_label     = BLANK + FormatPassthroughValue(BLANK);
  right2.m_sides[EAST].m_label      = std::string(BLANK) + "*";
  right2.m_sides[SOUTH].m_label     = std::string(BLANK) + "*";
  right2.m_sides[WEST].m_label      = COPYRIGHT;
  ts.push_back(right2);

  if (!bQuadOptimal) {
    Tile right3("RIGHT2");
    right3.m_display                  = BLANK;
    right3.m_sides[NORTH].m_strength  = 1;
    right3.m_sides[EAST].m_strength   = 1;
    right3.m_sides[SOUTH].m_strength  = 0;
    right3.m_sides[WEST].m_strength   = 2;
    right3.m_sides[NORTH].m_label     = std::string(BLANK) + "*";
    right3.m_sides[EAST].m_label      = "END@";
    right3.m_sides[SOUTH].m_label     = NODISPLAY;
    right3.m_sides[WEST].m_label      = std::string(BLANK) + "*";
    ts.push_back(right3);
  } else {
    Tile right3("RIGHT2");
    right3.m_display                  = BLANK;
    right3.m_sides[NORTH].m_strength  = 1;
    right3.m_sides[EAST].m_strength   = 1;
    right3.m_sides[SOUTH].m_strength  = 1;
    right3.m_sides[WEST].m_strength   = 2;
    right3.m_sides[NORTH].m_label     = std::string(BLANK) + "*";
    right3.m_sides[EAST].m_label      = "DROP";
    right3.m_sides[SOUTH].m_label     = "DROP";
    right3.m_sides[WEST].m_label      = std::string(BLANK) + "*";
    ts.push_back(right3);
  }



  //////////////////////////////////////////////////////////////////////
  //  Modify the tiletypes that simulate the TM to pass the binary
  //  counter values through from south to north.

  for (size_t i=0; i<tsTM.size(); i++) {
    int c=0;
    while (ctrVals[c].length() > 0) {
      Tile t = tsTM[i];
      std::string pv = FormatPassthroughValue(ctrVals[c].c_str());
      t.m_display += pv;
      t.m_sides[NORTH].m_label += pv;
      t.m_sides[SOUTH].m_label += pv;
      ts.push_back(t);
      c++;
    }
  }

  //////////////////////////////////////////////////////////////////////
  //  Create the tiles for the first row of a new TM computation
  int ct = 0;
  while (MostSigArray[ct].length() > 0) {
    int bit = GetCounterBit(MostSigArray[ct]);

    //  Start state
    std::string name = std::string("[") + TM.m_startState + "]," + itos(bit);
    Tile b;
    b.m_display                 = name;
    b.m_sides[NORTH].m_strength = 2;
    b.m_sides[EAST].m_strength  = 1;
    b.m_sides[SOUTH].m_strength = 2;
    b.m_sides[WEST].m_strength  = 0;
    b.m_sides[NORTH].m_label    = name + FormatPassthroughValue(MostSigArray[ct].c_str());
    b.m_sides[EAST].m_label     = COPYRIGHT;
    b.m_sides[SOUTH].m_label    = WrapCounterValue(MostSigArray[ct]);
    b.m_sides[WEST].m_label     = NODISPLAY;

    //  Special case for TM's that instantly halt
    for (size_t t=0; t<TM.m_transitions.size(); t++) {
      if ((TM.m_transitions[t].m_startState == TM.m_startState) &&
          (TM.m_transitions[t].m_input == itos(bit)))
      {
        if (TM.m_transitions[t].m_endState == TM.m_acceptingState) {
          b.m_sides[NORTH].m_label = ACCEPT + FormatPassthroughValue(MostSigArray[ct].c_str());
        } else if (TM.m_transitions[t].m_endState == TM.m_rejectingState) {
          b.m_sides[NORTH].m_label = REJECT + FormatPassthroughValue(MostSigArray[ct].c_str());
        }
      }
    }

    ts.push_back(b);
    ct++;
  }

  ct = 0;
  while (NonMostSigArray[ct].length() > 0) {
    int bit = GetCounterBit(NonMostSigArray[ct]);

    std::string name = itos(bit);
    Tile b;
    b.m_display                 = name;
    b.m_sides[NORTH].m_strength = 1;
    b.m_sides[EAST].m_strength  = 1;
    b.m_sides[SOUTH].m_strength = 1;
    b.m_sides[WEST].m_strength  = 1;
    b.m_sides[NORTH].m_label    = name + FormatPassthroughValue(NonMostSigArray[ct].c_str());
    b.m_sides[EAST].m_label     = COPYRIGHT;
    b.m_sides[SOUTH].m_label    = WrapCounterValue(NonMostSigArray[ct]);
    b.m_sides[WEST].m_label     = COPYRIGHT;
    ts.push_back(b);
    ct++;
  }

  Tile b;
  b.m_display                 = "_";
  b.m_sides[NORTH].m_strength = 1;
  b.m_sides[EAST].m_strength  = 1;
  b.m_sides[SOUTH].m_strength = 1;
  b.m_sides[WEST].m_strength  = 1;
  b.m_sides[NORTH].m_label    = std::string("_") + FormatPassthroughValue("_");
  b.m_sides[EAST].m_label     = COPYRIGHT;
  b.m_sides[SOUTH].m_label    = WrapCounterValue("_");
  b.m_sides[WEST].m_label     = COPYRIGHT;
  ts.push_back(b);

  b.m_display                 = "_*";
  b.m_sides[NORTH].m_strength = 1;
  b.m_sides[EAST].m_strength  = 2;
  b.m_sides[SOUTH].m_strength = 1;
  b.m_sides[WEST].m_strength  = 1;
  b.m_sides[NORTH].m_label    = std::string("_") + FormatPassthroughValue("_");
  b.m_sides[EAST].m_label     = "_*";
  b.m_sides[SOUTH].m_label    = WrapCounterValue("_*");
  b.m_sides[WEST].m_label     = COPYRIGHT;
  ts.push_back(b);

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  //  Create the tiles for the row that pass 'ACCEPT' or 'REJECT' out to the edge of the assembly

  //  Initial 'ACCEPT' or 'REJECT' tile for a computation
  int c=0;
  while (ctrVals[c].length() > 0) {

    if (ctrVals[c].find('*') != std::string::npos) {
      Tile a;
      a.m_display                 = std::string(ACCEPT) + FormatPassthroughValue(ctrVals[c].c_str());
      a.m_sides[NORTH].m_strength = 2;
      a.m_sides[EAST].m_strength  = 1;
      a.m_sides[SOUTH].m_strength = 2;
      a.m_sides[WEST].m_strength  = 1;
      a.m_sides[NORTH].m_label    = ctrVals[c];
      a.m_sides[EAST].m_label     = ACCEPT;
      a.m_sides[SOUTH].m_label    = std::string(ACCEPT) + FormatPassthroughValue(ctrVals[c].c_str());
      a.m_sides[WEST].m_label     = "CTR";
      a.m_color                   = ACCEPTCOLOR;
      ts.push_back(a);

      a.m_display                 = std::string(REJECT) + FormatPassthroughValue(ctrVals[c].c_str());
      a.m_sides[NORTH].m_strength = 2;
      a.m_sides[EAST].m_strength  = 1;
      a.m_sides[SOUTH].m_strength = 2;
      a.m_sides[WEST].m_strength  = 1;
      a.m_sides[NORTH].m_label    = ctrVals[c];
      a.m_sides[EAST].m_label     = REJECT;
      a.m_sides[SOUTH].m_label    = std::string(REJECT) + FormatPassthroughValue(ctrVals[c].c_str());
      a.m_sides[WEST].m_label     = "CTR";
      a.m_color                   = REJECTCOLOR;
      ts.push_back(a);
    } else {
      Tile a;
      a.m_display                 = std::string(ACCEPT) + FormatPassthroughValue(ctrVals[c].c_str());
      a.m_sides[NORTH].m_strength = 1;
      a.m_sides[EAST].m_strength  = 1;
      a.m_sides[SOUTH].m_strength = 2;
      a.m_sides[WEST].m_strength  = 1;
      a.m_sides[NORTH].m_label    = ctrVals[c];
      a.m_sides[EAST].m_label     = ACCEPT;
      a.m_sides[SOUTH].m_label    = std::string(ACCEPT) + FormatPassthroughValue(ctrVals[c].c_str());
      a.m_sides[WEST].m_label     = "CTR";
      a.m_color                   = ACCEPTCOLOR;
      ts.push_back(a);

      a.m_display                 = std::string(REJECT) + FormatPassthroughValue(ctrVals[c].c_str());
      a.m_sides[NORTH].m_strength = 1;
      a.m_sides[EAST].m_strength  = 1;
      a.m_sides[SOUTH].m_strength = 2;
      a.m_sides[WEST].m_strength  = 1;
      a.m_sides[NORTH].m_label    = ctrVals[c];
      a.m_sides[EAST].m_label     = REJECT;
      a.m_sides[SOUTH].m_label    = std::string(REJECT) + FormatPassthroughValue(ctrVals[c].c_str());
      a.m_sides[WEST].m_label     = "CTR";
      a.m_color                   = REJECTCOLOR;
      ts.push_back(a);
    }

    //  Tiles to the right of the initial ACCEPT or REJECT tile
    for (size_t s=0; s<TM.m_alphabet.size(); s++) {
      if (ctrVals[c].find('*') != std::string::npos) {
        Tile a;
        a.m_display                 = std::string(ACCEPT) + FormatPassthroughValue(ctrVals[c].c_str());
        a.m_sides[NORTH].m_strength = 2;
        a.m_sides[EAST].m_strength  = 1;
        a.m_sides[SOUTH].m_strength = 1;
        a.m_sides[WEST].m_strength  = 1;
        a.m_sides[NORTH].m_label    = ctrVals[c];
        a.m_sides[EAST].m_label     = ACCEPT;
        a.m_sides[SOUTH].m_label    = TM.m_alphabet[s] + FormatPassthroughValue(ctrVals[c].c_str());
        a.m_sides[WEST].m_label     = ACCEPT;
        a.m_color                   = ACCEPTCOLOR;
        ts.push_back(a);

        a.m_display                 = std::string(REJECT) + FormatPassthroughValue(ctrVals[c].c_str());
        a.m_sides[NORTH].m_strength = 2;
        a.m_sides[EAST].m_strength  = 1;
        a.m_sides[SOUTH].m_strength = 1;
        a.m_sides[WEST].m_strength  = 1;
        a.m_sides[NORTH].m_label    = ctrVals[c];
        a.m_sides[EAST].m_label     = REJECT;
        a.m_sides[SOUTH].m_label    = TM.m_alphabet[s] + FormatPassthroughValue(ctrVals[c].c_str());
        a.m_sides[WEST].m_label     = REJECT;
        a.m_color                   = REJECTCOLOR;
        ts.push_back(a);
      } else {
        Tile a;
        a.m_display                 = std::string(ACCEPT) + FormatPassthroughValue(ctrVals[c].c_str());
        a.m_sides[NORTH].m_strength = 1;
        a.m_sides[EAST].m_strength  = 1;
        a.m_sides[SOUTH].m_strength = 1;
        a.m_sides[WEST].m_strength  = 1;
        a.m_sides[NORTH].m_label    = ctrVals[c];
        a.m_sides[EAST].m_label     = ACCEPT;
        a.m_sides[SOUTH].m_label    = TM.m_alphabet[s] + FormatPassthroughValue(ctrVals[c].c_str());
        a.m_sides[WEST].m_label     = ACCEPT;
        a.m_color                   = ACCEPTCOLOR;
        ts.push_back(a);

        a.m_display                 = std::string(REJECT) + FormatPassthroughValue(ctrVals[c].c_str());
        a.m_sides[NORTH].m_strength = 1;
        a.m_sides[EAST].m_strength  = 1;
        a.m_sides[SOUTH].m_strength = 1;
        a.m_sides[WEST].m_strength  = 1;
        a.m_sides[NORTH].m_label    = ctrVals[c];
        a.m_sides[EAST].m_label     = REJECT;
        a.m_sides[SOUTH].m_label    = TM.m_alphabet[s] + FormatPassthroughValue(ctrVals[c].c_str());
        a.m_sides[WEST].m_label     = REJECT;
        a.m_color                   = REJECTCOLOR;
        ts.push_back(a);
      }
    }
    c++;
  }

  //  Finish out the ACCEPT and REJECT rows to the right
  Tile a;
  a.m_display                 = std::string(ACCEPT) + "_*";
  a.m_sides[NORTH].m_strength = 1;
  a.m_sides[EAST].m_strength  = 2;
  a.m_sides[SOUTH].m_strength = 1;
  a.m_sides[WEST].m_strength  = 1;
  a.m_sides[NORTH].m_label    = "_";
  a.m_sides[EAST].m_label     = std::string(ACCEPT) + "!";
  a.m_sides[SOUTH].m_label    = "_*";
  a.m_sides[WEST].m_label     = ACCEPT;
  a.m_color                   = ACCEPTCOLOR;
  ts.push_back(a);

  a.m_display                 = std::string(REJECT) + "_*";
  a.m_sides[NORTH].m_strength = 1;
  a.m_sides[EAST].m_strength  = 2;
  a.m_sides[SOUTH].m_strength = 1;
  a.m_sides[WEST].m_strength  = 1;
  a.m_sides[NORTH].m_label    = "_";
  a.m_sides[EAST].m_label     = std::string(REJECT) + "!";
  a.m_sides[SOUTH].m_label    = "_*";
  a.m_sides[WEST].m_label     = REJECT;
  a.m_color                   = REJECTCOLOR;
  ts.push_back(a);

  a.m_display                 = std::string(ACCEPT);
  a.m_sides[NORTH].m_strength = 1;
  a.m_sides[EAST].m_strength  = 1;
  a.m_sides[SOUTH].m_strength = 1;
  a.m_sides[WEST].m_strength  = 2;
  a.m_sides[NORTH].m_label    = "_*";
  if (!bQuadOptimal) {
    a.m_sides[EAST].m_label   = "END";
  } else {
    a.m_sides[EAST].m_label   = "DROP";
  }
  a.m_sides[SOUTH].m_label    = ACCEPT;
  a.m_sides[WEST].m_label     = std::string(ACCEPT) + "!";
  a.m_color                   = ACCEPTCOLOR;
  ts.push_back(a);

  a.m_display                 = std::string(REJECT);
  a.m_sides[NORTH].m_strength = 1;
  a.m_sides[EAST].m_strength  = 1;
  a.m_sides[SOUTH].m_strength = 1;
  a.m_sides[WEST].m_strength  = 2;
  a.m_sides[NORTH].m_label    = "_*";
  if (!bQuadOptimal) {
    a.m_sides[EAST].m_label   = "END";
  } else {
    a.m_sides[EAST].m_label   = "DROP";
  }
  a.m_sides[SOUTH].m_label    = REJECT;
  a.m_sides[WEST].m_label     = std::string(REJECT) + "!";
  a.m_color                   = REJECTCOLOR;
  ts.push_back(a);

  //  Finish out the CTR row to the left
  ct = 0;
  while (ctrVals[ct].length() > 0) {
    for (size_t i=0; i<TM.m_alphabet.size(); i++) {
      Tile a;
      a.m_display                 = std::string("CTR") + FormatPassthroughValue(ctrVals[ct].c_str());
      a.m_sides[NORTH].m_strength = 1;
      a.m_sides[EAST].m_strength  = 1;
      a.m_sides[SOUTH].m_strength = 1;
      a.m_sides[WEST].m_strength  = 1;
      a.m_sides[NORTH].m_label    = ctrVals[ct];
      a.m_sides[EAST].m_label     = "CTR";
      a.m_sides[SOUTH].m_label    = TM.m_alphabet[i] + FormatPassthroughValue(ctrVals[ct].c_str());
      a.m_sides[WEST].m_label     = "CTR";

      if (ctrVals[ct].find('*') != std::string::npos) {
        a.m_sides[NORTH].m_strength = 2;
      }
      ts.push_back(a);
    }
    ct++;
  }

  //////////////////////////////////////////////////////////////////////
  //  Create the seed tile
  Tile seed("SEED");
  seed.m_display                 = "SEED";
  seed.m_sides[NORTH].m_strength = 2;
  seed.m_sides[EAST].m_strength  = 2;
  seed.m_sides[SOUTH].m_strength = 1;
  seed.m_sides[WEST].m_strength  = 0;
  seed.m_sides[NORTH].m_label    = WrapCounterValue("M0*L");
  seed.m_sides[EAST].m_label     = "SEEDE";
  seed.m_sides[SOUTH].m_label    = "SOLN";
  seed.m_sides[WEST].m_label     = NODISPLAY;
  ts.push_back(seed);

  Tile seede("SEEDE");
  seede.m_display                 = "SEEDE";
  seede.m_sides[NORTH].m_strength = 1;
  seede.m_sides[EAST].m_strength  = 1;
  seede.m_sides[SOUTH].m_strength = 1;
  seede.m_sides[WEST].m_strength  = 2;
  seede.m_sides[NORTH].m_label    = WrapCounterValue("_*");
  seede.m_sides[EAST].m_label     = "DIAG";
  seede.m_sides[SOUTH].m_label    = "PRESOLN";
  seede.m_sides[WEST].m_label     = "SEEDE";
  ts.push_back(seede);

  //////////////////////////////////////////////////////////////////////
  //  Create a set of tiles for a regular counter to do the actual counting
  TileSet counterSet;
  GenerateCounterTileSet(2, NORTH, EAST, COUNTER_LOGWIDTH, &counterSet, false);

  if (bQuadOptimal) {
    for (std::vector<Tile>::iterator it = counterSet.begin(); it != counterSet.end(); it++) {
      if (it->m_sides[SOUTH].m_label == "M1*") {
        it->m_display                 = "M1";
        it->m_sides[NORTH].m_strength = 2;
        it->m_sides[EAST].m_strength  = 1;
        it->m_sides[SOUTH].m_strength = 2;
        it->m_sides[WEST].m_strength  = 0;
        it->m_sides[NORTH].m_label    = "M1";
        it->m_sides[EAST].m_label     = "$";
        it->m_sides[SOUTH].m_label    = "M1*";
        it->m_sides[WEST].m_label     = NODISPLAY;

      } else if (it->m_sides[SOUTH].m_label == "M1*L") {
        it->m_display                 = "M1";
        it->m_sides[NORTH].m_strength = 2;
        it->m_sides[EAST].m_strength  = 1;
        it->m_sides[SOUTH].m_strength = 2;
        it->m_sides[WEST].m_strength  = 0;
        it->m_sides[NORTH].m_label    = "M1";
        it->m_sides[EAST].m_label     = "$L";
        it->m_sides[SOUTH].m_label    = "M1*L";
        it->m_sides[WEST].m_label     = NODISPLAY;
      }
    }

    Tile tc;
    tc.m_display                 = "0";
    tc.m_sides[NORTH].m_strength = 1;
    tc.m_sides[EAST].m_strength  = 1;
    tc.m_sides[SOUTH].m_strength = 1;
    tc.m_sides[WEST].m_strength  = 1;
    tc.m_sides[NORTH].m_label    = "0";
    tc.m_sides[EAST].m_label     = "$L";
    tc.m_sides[SOUTH].m_label    = "1L";
    tc.m_sides[WEST].m_label     = "$";
    counterSet.push_back(tc);

    tc.m_display                 = "0";
    tc.m_sides[NORTH].m_strength = 1;
    tc.m_sides[EAST].m_strength  = 1;
    tc.m_sides[SOUTH].m_strength = 1;
    tc.m_sides[WEST].m_strength  = 1;
    tc.m_sides[NORTH].m_label    = "0";
    tc.m_sides[EAST].m_label     = "$";
    tc.m_sides[SOUTH].m_label    = "1";
    tc.m_sides[WEST].m_label     = "$";
    counterSet.push_back(tc);

    tc.m_display                 = "0*L";
    tc.m_sides[NORTH].m_strength = 1;
    tc.m_sides[EAST].m_strength  = 1;
    tc.m_sides[SOUTH].m_strength = 1;
    tc.m_sides[WEST].m_strength  = 1;
    tc.m_sides[NORTH].m_label    = "0*L";
    tc.m_sides[EAST].m_label     = "__";
    tc.m_sides[SOUTH].m_label    = "_";
    tc.m_sides[WEST].m_label     = "$L";
    counterSet.push_back(tc);
  }

  for (size_t i=0; i<counterSet.size(); i++) {
    counterSet[i].m_sides[NORTH].m_label = WrapCounterValue(counterSet[i].m_sides[NORTH].m_label);

    //  Modify generic counter tiles so that:
    //    1.  Tiles can attach EAST of the LSB
    //    2.  The tile with the marker ('*') does not initiate upward growth
    //    3.  The MSB initiates the upward growth (since the TM tape head needs to start at the leftmost cell)
    if (counterSet[i].m_display.find('L') != std::string::npos) {
      counterSet[i].m_sides[EAST].m_strength  = 1;
      counterSet[i].m_sides[EAST].m_label     = "__";
    }

    if (counterSet[i].m_display.find('*') != std::string::npos) {
      counterSet[i].m_sides[NORTH].m_strength = 1;
    }

    if (counterSet[i].m_display.find('M') != std::string::npos) {
      counterSet[i].m_sides[NORTH].m_strength = 2;
    }
  }
  ts += counterSet;

  Tile x;
  x.m_display                 = "__";
  x.m_sides[NORTH].m_strength = 1;
  x.m_sides[EAST].m_strength  = 1;
  x.m_sides[SOUTH].m_strength = 1;
  x.m_sides[WEST].m_strength  = 1;
  x.m_sides[NORTH].m_label    = std::string("_") + FormatPassthroughValue("_");
  x.m_sides[EAST].m_label     = "__";
  x.m_sides[SOUTH].m_label    = "_";
  x.m_sides[WEST].m_label     = "__";
  ts.push_back(x);

  x.m_display                 = "__*";
  x.m_sides[NORTH].m_strength = 1;
  x.m_sides[EAST].m_strength  = 1;
  x.m_sides[SOUTH].m_strength = 1;
  x.m_sides[WEST].m_strength  = 1;
  x.m_sides[NORTH].m_label    = WrapCounterValue("_*");
  if (!bQuadOptimal) {
    x.m_sides[EAST].m_label   = "END";
  } else {
    x.m_sides[EAST].m_label   = "DROP";
  }
  x.m_sides[SOUTH].m_label    = "_*";
  x.m_sides[WEST].m_label     = "__";
  ts.push_back(x);

  if (!bQuadOptimal) {
    //////////////////////////////////////////////////////////////////////
    //  Create the tiles to creep down the right edge
    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = ACCEPT;
    x.m_sides[EAST].m_label     = "END";
    x.m_sides[SOUTH].m_label    = ACCEPT;
    x.m_sides[WEST].m_label     = "END";
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = REJECT;
    x.m_sides[EAST].m_label     = "END";
    x.m_sides[SOUTH].m_label    = REJECT;
    x.m_sides[WEST].m_label     = "END";
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 2;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = ACCEPT;
    x.m_sides[EAST].m_label     = "END";
    x.m_sides[SOUTH].m_label    = std::string(ACCEPT) + "!";
    x.m_sides[WEST].m_label     = "END@";
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 2;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = REJECT;
    x.m_sides[EAST].m_label     = "END";
    x.m_sides[SOUTH].m_label    = std::string(REJECT) + "!";
    x.m_sides[WEST].m_label     = "END@";
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 2;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = std::string(ACCEPT) + "!";
    x.m_sides[EAST].m_label     = "END@";
    x.m_sides[SOUTH].m_label    = "END";
    x.m_sides[WEST].m_label     = ACCEPT;
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 2;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = std::string(REJECT) + "!";
    x.m_sides[EAST].m_label     = "END@";
    x.m_sides[SOUTH].m_label    = "END";
    x.m_sides[WEST].m_label     = REJECT;
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 0;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 2;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = NODISPLAY;
    x.m_sides[EAST].m_label     = ACCEPT;
    x.m_sides[SOUTH].m_label    = std::string(ACCEPT) + "!";
    x.m_sides[WEST].m_label     = "END@";
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 0;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 2;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = NODISPLAY;
    x.m_sides[EAST].m_label     = REJECT;
    x.m_sides[SOUTH].m_label    = std::string(REJECT) + "!";
    x.m_sides[WEST].m_label     = "END@";
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 0;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = NODISPLAY;
    x.m_sides[EAST].m_label     = ACCEPT;
    x.m_sides[SOUTH].m_label    = ACCEPT;
    x.m_sides[WEST].m_label     = "END";
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 0;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = NODISPLAY;
    x.m_sides[EAST].m_label     = REJECT;
    x.m_sides[SOUTH].m_label    = REJECT;
    x.m_sides[WEST].m_label     = "END";
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    //  Start to 'straighten out' the results and bring them to the y-axis
    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = "PRESOLN";
    x.m_sides[EAST].m_label     = ACCEPT;
    x.m_sides[SOUTH].m_label    = "PRESOLN";
    x.m_sides[WEST].m_label     = ACCEPT;
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = "PRESOLN";
    x.m_sides[EAST].m_label     = REJECT;
    x.m_sides[SOUTH].m_label    = "PRESOLN";
    x.m_sides[WEST].m_label     = REJECT;
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 0;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 2;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = NODISPLAY;
    x.m_sides[EAST].m_label     = ACCEPT;
    x.m_sides[SOUTH].m_label    = std::string("DIAG") + ACCEPT;
    x.m_sides[WEST].m_label     = "DIAG";
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 0;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 2;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = NODISPLAY;
    x.m_sides[EAST].m_label     = REJECT;
    x.m_sides[SOUTH].m_label    = std::string("DIAG") + REJECT;
    x.m_sides[WEST].m_label     = "DIAG";
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 2;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = std::string("DIAG") + ACCEPT;
    x.m_sides[EAST].m_label     = "DIAG";
    x.m_sides[SOUTH].m_label    = "PRESOLN";
    x.m_sides[WEST].m_label     = ACCEPT;
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 2;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = std::string("DIAG") + REJECT;
    x.m_sides[EAST].m_label     = "DIAG";
    x.m_sides[SOUTH].m_label    = "PRESOLN";
    x.m_sides[WEST].m_label     = REJECT;
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    x.m_display                 = ACCEPT;
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 0;
    x.m_sides[NORTH].m_label    = "SOLN";
    x.m_sides[EAST].m_label     = ACCEPT;
    x.m_sides[SOUTH].m_label    = "SOLN";
    x.m_sides[WEST].m_label     = NODISPLAY;
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = REJECT;
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 0;
    x.m_sides[NORTH].m_label    = "SOLN";
    x.m_sides[EAST].m_label     = REJECT;
    x.m_sides[SOUTH].m_label    = "SOLN";
    x.m_sides[WEST].m_label     = NODISPLAY;
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);
  } else {
    //  Tiles for dropping down answers of computations for quadrant optimal assemblies
    x.m_display                 = "DROP";
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = "DROP";
    x.m_sides[EAST].m_label     = "DROP";
    x.m_sides[SOUTH].m_label    = "DROP";
    x.m_sides[WEST].m_label     = "DROP";
    ts.push_back(x);

    x.m_display                 = "DROP";
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 0;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = "DROP";
    x.m_sides[EAST].m_label     = "DIAG";
    x.m_sides[SOUTH].m_label    = NODISPLAY;
    x.m_sides[WEST].m_label     = "DIAG";
    ts.push_back(x);

    x.m_display                 = "ACCEPT";
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = "ACCEPT";
    x.m_sides[EAST].m_label     = "DROP";
    x.m_sides[SOUTH].m_label    = "ACCEPT";
    x.m_sides[WEST].m_label     = "DROP";
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = "REJECT";
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 1;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = "REJECT";
    x.m_sides[EAST].m_label     = "DROP";
    x.m_sides[SOUTH].m_label    = "REJECT";
    x.m_sides[WEST].m_label     = "DROP";
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);

    x.m_display                 = "ACCEPT";
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 0;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = "ACCEPT";
    x.m_sides[EAST].m_label     = "DIAG";
    x.m_sides[SOUTH].m_label    = NODISPLAY;
    x.m_sides[WEST].m_label     = "DIAG";
    x.m_color                   = ACCEPTCOLOR;
    ts.push_back(x);

    x.m_display                 = "REJECT";
    x.m_sides[NORTH].m_strength = 1;
    x.m_sides[EAST].m_strength  = 1;
    x.m_sides[SOUTH].m_strength = 0;
    x.m_sides[WEST].m_strength  = 1;
    x.m_sides[NORTH].m_label    = "REJECT";
    x.m_sides[EAST].m_label     = "DIAG";
    x.m_sides[SOUTH].m_label    = NODISPLAY;
    x.m_sides[WEST].m_label     = "DIAG";
    x.m_color                   = REJECTCOLOR;
    ts.push_back(x);
  }


  //////////////////////////////////////////////////////////////////////
  //  Rotate the tileset so that the results will be enumerated on the x-axis
  for (size_t i=0; i<ts.size(); i++) {
    if (bDoRotate) {
      ts[i].Rotate(270);
    }
  }

  //////////////////////////////////////////////////////////////////////
  //  Output the tileset

  std::ofstream out(TileOutFile.c_str());
  if (!out.is_open()) {
    printf("Error:  Failed to open the output file '%s'\n", TileOutFile.c_str());
    exit(1);
  }

  printf("\nOutputting tileset to file: '%s'...\n\n", TileOutFile.c_str());
  for (size_t j=0; j<ts.size(); j++) {
    ts[j].Print(&out);
    out << "\n";
  }
  out.close();


  //////////////////////////////////////////////////////////////////////
  //  Output the seed definition

  std::ofstream seedOut(SeedOutFile.c_str());
  if (!seedOut.is_open()) {
    printf("Error:  Failed to open the seed output file '%s'\n", SeedOutFile);
    exit(1);
  }

  printf("Outputting seed definition to file: '%s'...\n\n", SeedOutFile.c_str());

  std::string tdpFile = TileOutFile.substr((TileOutFile.find_last_of('\\')+1), TileOutFile.length());
  seedOut << tdpFile;
  seedOut << "\n";
  if (bQuadOptimal) {
    seedOut << "SEED 0 0\n";
  } else if (bDoRotate) {
    seedOut << "SEED -1 0\n";
  } else {
    seedOut << "SEED 0 1\n";
  }

  printf("DeciderTiler has completed successfully.\n");
	return 0;
}

void PrintTemplate(const char* tmplFileName)
{
  if (!tmplFileName || !_tcslen(tmplFileName)) {
    printf("\nError:  No output file was specified for the template!\n");
    exit(1);
  }

  std::ofstream out(tmplFileName);
  if (!out.is_open()) {
    printf("Error:  Failed to open the template output file '%s'\n", tmplFileName);
    exit(1);
  }

  printf("\nOutputting template to file: '%s'...\n\n", tmplFileName);

  out << "/////////////////////////////////////////////////////////////////////////////\n";
  out << "//  Turing Machine Specification\n";
  out << "//    Assumptions:\n";
  out << "//      - the tape is one-way infinite to the right\n";
  out << "//      - the input alphabet is {0, 1} and the Turing machine\n";
  out << "//        accepts as input the binary representation of each\n";
  out << "//        natural number\n";
  out << "//      - the Turing machine halts on every input, either in the\n";
  out << "//        accepting state or the rejecting state\n";
  out << "//      - the TM will begin in the start state with the tape head\n";
  out << "//        on the leftmost cell of the tape, which will be the most\n";
  out << "//        significant bit of the input binary number\n";
  out << "//      - the input alphabet is {0, 1}\n";
  out << "//      - the tape alphabet will automatically include {0, 1, _}\n";
  out << "//        where the '_' symbol is used to designate blanks\n";
  out << "//    Reserved characters:\n";
  out << "//      - These symbols are used by DeciderTiler in generated names and\n";
  out << "//        therefore can't appear in the input alphabet or state names:\n";
  out << "//          ";
  int i = 0;
  while (ReservedChars[i]) {
    if (i > 0) {
      out << ", ";
    }
    out << std::string("'") + ReservedChars[i] + std::string("'");
    i++;
  }
  out << "\n";
  out << "//    Naming convention:\n";
  out << "//      - DeciderTiler will automatically generate tile names for most\n";
  out << "//        tiles as a concatention of four edge labels as follows:\n";
  out << "//          <north label>|<east lable>|<south label>|<west label>\n";
  out << "/////////////////////////////////////////////////////////////////////////////\n";
  out << "\n";
  out << "//  States:  a comma separated list of state names (e.g. \"q0, q1, q2\")\n";
  out << "States: \n";
  out << "\n";
  out << "//  Alphabet:  a comma separated list of tape alphabet symbols, excluding \"0\",\n";
  out << "//   \"1\", and \"_\" which are included automatically (e.g. \"2, 3, 4\")\n";
  out << "Alphabet: \n";
  out << "\n";
  out << "//  Start state: the name of the state to be treated as the start state\n";
  out << "Start state: \n";
  out << "\n";
  out << "//  Accepting state:  the name of the state to be treated as the accepting state (e.g. \"q5\")\n";
  out << "Accepting state: \n";
  out << "\n";
  out << "//  Rejecting state:  the name of the state to be treated as the rejecting state (e.g. \"q6\")\n";
  out << "Rejecting state: \n";
  out << "\n";
  out << "//  Transitions:  a list of transitions, one per line, in the following form\n";
  out << "//    <state>, <input> -> <state>, <output>, <direction: L|R>\n";
  out << "//    Example:  q0, 0 -> q1, 1, R\n";
  out << "Transitions: (start on next line)\n";
  out << "\n";
  out << "\n";

  out.close();
  printf("Completed successfully.\n");
}
